from pathlib import Path

import msgspec

from phylodata.data_types import EditablePaperWithExperiment
from phylodata.process.samples.taxon_classification import look_up_taxon_classification


def change_ncbi(metadata_file: Path, sample_id: str, ncbi_taxon_id: int):
    """
    Changes the classification of a NCBI sample in a editable metadata file.

    Args:
        metadata_file: The path to the metadata file.
        sample_id: The ID of the sample to correct.
        ncbi_taxon_id: The NCBI taxon ID.
    """

    with open(metadata_file, "r") as file:
        metadata = msgspec.json.decode(file.read(), type=EditablePaperWithExperiment)

    for sample in metadata.samples:
        if sample.sample_id == sample_id:
            if classification := look_up_taxon_classification(ncbi_taxon_id):
                sample.classification = classification
                sample.scientific_name = classification[0].scientific_name
                sample.common_name = classification[0].common_name
                break

    with open(metadata_file, "wb") as file:
        file.write(msgspec.json.format(msgspec.json.encode(metadata), indent=2))



import os


def add_file_name_suffix(file_name: str, suffix: str, delimiter: str = "") -> str:
    """Appends the suffix to the file name while preserving
    the file extension."""
    base, ext = os.path.splitext(file_name)
    return f"{base}{delimiter}{suffix}{ext}"



import re

DOI_REGEX = r"^https?:\/\/(?:www\.)?(?:dx\.)?doi\.org\/10\.\d{4,}\/[^\s]+$"


def is_doi(doi: str) -> bool:
    """Checks whether the given string is a valid DOI URL."""
    pattern = re.compile(DOI_REGEX)
    return bool(pattern.match(doi))


def get_doi_url(doi: str) -> str:
    """Returns the URL for the given DOI."""
    if is_doi(doi):
        return doi
    else:
        return f"https://doi.org/{doi}"



from datetime import datetime

import streamlit as st

from phylodata.data_types import EditablePaper, NonEditablePaper
from phylodata.errors import ValidationError
from phylodata.process.module import Module
from phylodata.process.paper.bibtex import (
    get_bibtex_value,
    is_valid_bibtex,
    yield_bibtex_authors,
)
from phylodata.process.paper.doi import get_doi_url, is_doi
from phylodata.process.paper.mail import is_valid_email


class PaperModule(Module[tuple[EditablePaper, NonEditablePaper]]):
    def fill_using_bibtex(self):
        if "paper_bibtex_title" in st.session_state:
            # we only set it the first time to avoid overwriting manual corrections
            return

        bibtex: str = st.session_state["paper_bibtex"]

        if name := get_bibtex_value(bibtex, "title"):
            st.session_state["paper_bibtex_title"] = name
        if abstract := get_bibtex_value(bibtex, "abstract"):
            st.session_state["paper_bibtex_abstract"] = abstract
        if (year := get_bibtex_value(bibtex, "year")) and year.isdigit():
            st.session_state["paper_bibtex_year"] = int(year)
        if authors := get_bibtex_value(bibtex, "author"):
            st.session_state["paper_bibtex_authors"] = "\n".join(
                yield_bibtex_authors(authors)
            )
        if doi := get_bibtex_value(bibtex, "doi"):
            st.session_state["paper_bibtex_doi"] = get_doi_url(doi)

    def ui(self):
        with st.container(border=True):
            st.write("""
            ### Paper

            Enter the following general information about your paper:
            """)
            self.bibtex = st.text_area(
                "BibTex citation",
                on_change=self.fill_using_bibtex,
                key="paper_bibtex",
                height=200,
            )
            self.title = st.text_input(
                "Title", value=st.session_state.get("paper_bibtex_title")
            )
            self.year = st.number_input(
                "Publication year",
                step=1,
                value=st.session_state.get("paper_bibtex_year"),
            )
            self.abstract = st.text_area(
                "Abstract", value=st.session_state.get("paper_bibtex_abstract")
            )
            self.authors = st.text_area(
                "Authors", value=st.session_state.get("paper_bibtex_authors")
            )
            st.markdown(
                """One author per line. Use *Name Surname* like *Joseph Felsenstein*."""
            )
            self.doi = st.text_input(
                "DOI", value=st.session_state.get("paper_bibtex_doi")
            )
            st.markdown(
                """Use the URL format (like *https://doi.org/10.1093/sysbio/22.3.240*).
                Experiments with the same paper DOI will be linked to the same paper."""
            )
            self.email = st.text_input("E-Mail")
            st.markdown(
                """We need your E-Mail in order to contact you if there are any issues with the experiment."""
            )
            self.url = st.text_input("URL (optional)")

    def validate(self):
        if not self.title or not self.title.strip():
            raise ValidationError("Specify a paper title.")
        if not self.year or self.year < 1800 or datetime.now().year + 1 < self.year:
            raise ValidationError("Specify a valid paper publication year.")
        if not self.abstract or not self.abstract.strip():
            raise ValidationError("Specify a paper abstract.")
        if not self.authors or not self.authors.strip():
            raise ValidationError("Specify at least one paper author.")
        if not is_valid_bibtex(self.bibtex):
            raise ValidationError("Specify exactly one bibtex entry.")
        if not is_valid_email(self.email.strip()):
            raise ValidationError("Specify a valid email address.")
        if not self.doi or not is_doi(self.doi.strip()):
            raise ValidationError("Specify a DOI (like https://doi.org/10.1000/182).")

    def parse(self) -> tuple[EditablePaper, NonEditablePaper]:
        if (
            not self.title
            or not self.year
            or not self.authors
            or not self.abstract
            or not self.doi
        ):
            raise ValidationError("Paper info missing.")

        return EditablePaper(
            title=self.title.strip(),
            year=int(self.year),
            authors=[author.strip() for author in self.authors.split("\n")],
            abstract=self.abstract.strip(),
            bibtex=self.bibtex.strip(),
            url=self.url.strip() or None,
            email=self.email.strip(),
        ), NonEditablePaper(doi=self.doi.strip())



import csv
from io import BytesIO, TextIOWrapper
from math import floor
from typing import Generator, Optional

from commonnexus.blocks.trees import Trees

from phylodata.data_types import File, FileType
from phylodata.errors import ValidationError
from phylodata.process.utils.bytesio_utils import (
    get_nexus_from_bytesio,
    get_xml_from_bytesio,
)
from phylodata.process.utils.file_utils import add_file_name_suffix

MIN_NUM_SNAPSHOTS = 50
"""The minimum number of snapshots required for a valid BEAST 2 posterior log file.
When the number of snapshots is less than this value, the log file is considered invalid."""

PREVIEW_FRACTION = 0.05
"""The fraction of state snapshots that are stored in the preview files."""


def parse_file(
    file: BytesIO, file_type: Optional[FileType] = FileType.UNKNOWN
) -> Generator[File, None, None]:
    match file_type:
        case FileType.BEAST2_CONFIGURATION:
            yield from parse_beast2_config(file)
        case FileType.BEAST2_POSTERIOR_LOGS:
            yield from parse_beast2_logs(file)
        case FileType.POSTERIOR_TREES:
            yield from parse_beast2_trees(file)
        case _:
            yield from parse_other_file(file)


def parse_beast2_config(file: BytesIO) -> Generator[File, None, None]:
    """Parses a BEAST 2 config file. If the file does not seem to be a valid
    BEAST config, a ValidationError is raised."""
    xml = get_xml_from_bytesio(file)
    root = xml.getroot()

    if root is None or root.tag.lower() != "beast":
        raise ValidationError("BEAST 2 configuration has no root BEAST tag.")

    if not root.attrib.get("version", "").startswith("2."):
        raise ValidationError(
            "The BEAST configuration file is not for BEAST 2 (likely, it is for BEAST X)."
        )

    if not any(child.tag.lower() == "data" for child in root) and not any(
        child.tag.lower() == "alignment" for child in root
    ):
        raise ValidationError("BEAST 2 has no <data> tag.")

    if not any(child.tag.lower() == "run" for child in root):
        raise ValidationError("BEAST 2 has no <run> tag.")

    yield File.from_bytes(
        file,
        name=file.name,
        type=FileType.BEAST2_CONFIGURATION,
    )


def parse_beast2_logs(file: BytesIO) -> Generator[File, None, None]:
    """Parses a BEAST 2 log file. If the file does not seem to be a valid
    BEAST log, a ValidationError is raised."""
    try:
        wrapper = TextIOWrapper(file)

        def uncomment_lines(wrapper: TextIOWrapper):
            for line in wrapper:
                if not line.lstrip().startswith("#"):
                    yield line

        tsv_file = csv.DictReader(uncomment_lines(wrapper), delimiter="\t")
    except Exception:
        raise ValidationError("BEAST 2 log file is not a tab-separated file.")

    rows = list(tsv_file)
    num_rows = len(rows)

    if num_rows < MIN_NUM_SNAPSHOTS:
        raise ValidationError(
            f"BEAST 2 log file should contain at least {MIN_NUM_SNAPSHOTS} entries."
        )

    # detach wrapper to avoid closing the BytesIO when wrapper is garbage collected
    wrapper.detach()

    # yield full file

    yield File.from_bytes(
        file,
        name=file.name,
        type=FileType.BEAST2_POSTERIOR_LOGS,
    )

    # generate preview file

    preview_lines = rows[: floor(len(rows) * PREVIEW_FRACTION)]

    preview_file = BytesIO()
    preview_wrapper = TextIOWrapper(preview_file, encoding="utf-8", newline="")

    preview_writer = csv.DictWriter(
        preview_wrapper, fieldnames=preview_lines[0].keys(), delimiter="\t"
    )
    preview_writer.writeheader()
    preview_writer.writerows(preview_lines)

    # detach wrapper to avoid closing the BytesIO when wrapper is garbage collected

    preview_wrapper.detach()

    # yield preview file

    yield File.from_bytes(
        preview_file,
        name=add_file_name_suffix(file.name, " (preview)"),
        type=FileType.BEAST2_POSTERIOR_LOGS,
        is_preview=True,
    )


def parse_beast2_trees(file: BytesIO) -> Generator[File, None, None]:
    """Parses a BEAST 2 trees file. If the file does not seem to be a valid
    BEAST trees file, a ValidationError is raised."""
    nexus = get_nexus_from_bytesio(file)

    if not nexus.TREES or not nexus.TREES.trees:
        raise ValidationError("BEAST 2 trees contains no actual trees.")

    num_trees = len(nexus.TREES.trees)

    if num_trees < MIN_NUM_SNAPSHOTS:
        raise ValidationError(
            f"BEAST 2 trees should contain at least {MIN_NUM_SNAPSHOTS} trees."
        )

    # yield full trees file

    yield File.from_bytes(
        file,
        name=file.name,
        type=FileType.POSTERIOR_TREES,
    )

    # generate preview file

    preview_trees_commands = []
    num_previewed_trees = 0

    for command in nexus.TREES:
        # test if it is a TREE command
        if command.name.lower() == "tree":
            if num_previewed_trees < num_trees * PREVIEW_FRACTION:
                preview_trees_commands.append(command)
                num_previewed_trees += 1
        else:
            # this is not a tree command
            # we keep it
            preview_trees_commands.append(command)

    nexus.replace_block(nexus.TREES, Trees(nexus, preview_trees_commands))

    preview_file = BytesIO()
    preview_file.write(str(nexus).encode("utf-8"))

    # yield preview file

    yield File.from_bytes(
        preview_file,
        name=add_file_name_suffix(file.name, " (preview)"),
        type=FileType.POSTERIOR_TREES,
        is_preview=True,
    )


def parse_other_file(file: BytesIO) -> Generator[File, None, None]:
    # if the file contains a single tree => summary tree

    try:
        nexus = get_nexus_from_bytesio(file)
        if nexus.TREES and nexus.TREES.trees and len(nexus.TREES.trees) == 1:
            yield File.from_bytes(
                file,
                name=file.name,
                type=FileType.SUMMARY_TREE,
            )
            return
    except ValidationError:
        # file is not a valid trees file
        ...

    # otherwise we yield an UNKNOWN file

    yield File.from_bytes(
        file,
        name=file.name,
        type=FileType.UNKNOWN,
    )



import itertools
from io import BytesIO
from xml.etree import ElementTree

from commonnexus import Nexus

from phylodata.errors import ValidationError


def get_nexus_from_bytesio(file: BytesIO) -> Nexus:
    file.seek(0)
    try:
        lines = itertools.chain.from_iterable(
            (string_line.decode("utf-8") for string_line in file)
        )
        return Nexus(lines)
    except Exception:
        raise ValidationError("BEAST 2 trees file is no valid NEXUS file.")


def get_xml_from_bytesio(file: BytesIO) -> ElementTree.ElementTree:
    file.seek(0)
    try:
        return ElementTree.parse(file)  # type: ignore
    except Exception as e:
        raise ValidationError(f"BEAST 2 configuration is no valid XML file. ({e})")



from pathlib import Path
from typing import Optional

import requests


def download_file(
    directory: Path,
    experiment: str,
    file_name: str,
    version: Optional[int] = None,
    force_download: Optional[bool] = False,
) -> Path:
    """Downloads an experiment file and stores it in the given directory. If no version is given,
    the most recent one is downloaded.

    Raises a ValueError if the file could not be found."""
    downloaded_file = directory / file_name

    if not force_download and downloaded_file.exists():
        return downloaded_file

    if version:
        download_link = requests.get(
            f"https://phylodata.com/api/getDownloadLink/{experiment}/{file_name}/{version}"
        ).text
    else:
        download_link = requests.get(
            f"https://phylodata.com/api/getDownloadLink/{experiment}/{file_name}"
        ).text

    response = requests.get(
        download_link, headers={"content-type": "multipart/form-data"}
    )

    if response.status_code == 404:
        raise ValueError("Unknown experiment or file.")

    with open(downloaded_file, mode="wb") as file:
        file.write(response.content)

    return downloaded_file



from loguru import logger

from phylodata.data_types import ClassificationEntry, DataType, Sample
from phylodata.errors import BlastError
from phylodata.process.samples.run_blast import extract_taxon_ids, run_blast
from phylodata.process.samples.taxon_classification import look_up_taxon_classification

MAX_SEQ_LENGTH_CONSIDERED = 2500


def add_nucleotide_metadata(samples: list[Sample]) -> list[Sample]:
    """Tries to add metadata to all samples for which we have nucleotide
    sequences."""
    # collect all nucleotide sequences

    nucleotide_sequences: list[str] = []
    nucleotide_sequence_idx: list[int] = []

    for idx, sample in enumerate(samples):
        if sample.classification:
            continue

        for data in sample.sample_data:
            if data.type in [DataType.RNA, DataType.DNA]:
                nucleotide_sequences.append(
                    "".join(c for c in data.data if c.isalpha())
                )
                nucleotide_sequence_idx.append(idx)

    # fetch the nucleotide classifications

    classification = fetch_nucleotide_classifications(nucleotide_sequences)

    # add the classifications to the samples

    for idx, classification in zip(nucleotide_sequence_idx, classification):
        if classification:
            samples[idx].classification = classification
            samples[idx].scientific_name = classification[0].scientific_name
            samples[idx].common_name = classification[0].common_name

    return samples


def fetch_nucleotide_classifications(
    sequences: list[str],
) -> list[list[ClassificationEntry] | None]:
    if not sequences:
        return []

    blast_params = {
        "CMD": "Put",
        "PROGRAM": "blastn",
        "MEGABLAST": True,
        "DATABASE": "core_nt",
        "HITLIST_SIZE": 1,
    }

    try:
        blast_results = run_blast(sequences, blast_params, MAX_SEQ_LENGTH_CONSIDERED)
    except BlastError:
        return [None] * len(sequences)

    taxon_ids = extract_taxon_ids(blast_results)

    classifications: list[list[ClassificationEntry] | None] = []

    for taxon_id in taxon_ids:
        if taxon_id:
            try:
                classifications.append(look_up_taxon_classification(taxon_id))
            except Exception as e:
                logger.error(f"Taxon look up faild: {e}")
                classifications.append(None)
        else:
            classifications.append(None)

    return classifications



from collections import defaultdict
from io import BytesIO
from typing import Generator
from xml.etree import ElementTree

from phylodata.data_types import (
    DataType,
    Sample,
    SampleData,
)
from phylodata.process.samples.add_language_metadata import add_language_metadata
from phylodata.process.samples.add_nucleotide_metadata import add_nucleotide_metadata
from phylodata.process.samples.add_protein_metadata import add_protein_metadata
from phylodata.process.samples.sequence_utils import (
    contains_sequence,
    is_amino_acid_sequence,
    is_dna_sequence,
    is_rna_sequence,
)
from phylodata.process.utils.bytesio_utils import get_xml_from_bytesio


def parse_beast2_samples(beast2_config: BytesIO) -> list[Sample]:
    """Collects all sample data found in the BEAST 2 XML.

    All top-level <data> tags are traversed in order to find all
    alignments in the file.

    Note that this only works with basic BEAST XML files and will not
    work for every possible configuration."""
    sample_data_per_id: dict[str, list[SampleData]] = collect_sample_data(beast2_config)
    samples = construct_samples_from_data(sample_data_per_id)
    return samples


def collect_sample_data(beast2_config: BytesIO) -> dict[str, list[SampleData]]:
    """Collects all samples from the data tags in the given BEAST 2 XML."""
    xml = get_xml_from_bytesio(beast2_config)
    xml_root = xml.getroot()

    if xml_root is None:
        return {}

    # we first look at all <data> tags to find any sequence

    sample_data: list[tuple[str, SampleData]] = []
    for root_child in xml_root:
        if root_child.tag.lower() == "data":
            sample_data += collect_sample_data_from_data_tag(root_child)

    # group the samples by their id (taxon)

    sample_data_per_id: dict[str, list[SampleData]] = defaultdict(list)
    for sample_id, data in sample_data:
        sample_data_per_id[sample_id].append(data)

    return sample_data_per_id


def collect_sample_data_from_data_tag(
    data_element: ElementTree.Element,
) -> Generator[tuple[str, SampleData], None, None]:
    """Yields all sample data found in BEAST 2 XML <sequence> tags within
    a  given <data> tag."""
    data_type = data_element.attrib.get("dataType", "standard")

    match data_type:
        case "aminoacid":
            data_type = DataType.AMINO_ACIDS
        case "binary" | "integer":
            data_type = DataType.TRAITS
        case _:  # "nucleotide" | "standard" | "twoStateCovarion" | "user defined"
            data_type = DataType.UNKNOWN

    for sequence in data_element:
        if sequence.tag.lower() != "sequence":
            continue

        sample_id = sequence.attrib["taxon"]
        data = sequence.attrib.get("value") or sequence.text

        if not data:
            continue

        # ignore whitespace
        data = "".join(data.split())

        # data can be comma-delimited or just a joined string
        if "," in data:
            data = "".join(data.split(","))

        # if dataType was not given in the <data> tag, we try to infer it from
        # the sequence itself
        if data_type == DataType.UNKNOWN:
            if not contains_sequence(data):
                # we don't have any actual useful characters
                ...
            elif is_dna_sequence(data):
                data_type = DataType.DNA
            elif is_rna_sequence(data):
                data_type = DataType.RNA
            elif is_amino_acid_sequence(data):
                data_type = DataType.AMINO_ACIDS

        yield sample_id, SampleData(type=data_type, length=len(data), data=data)


def construct_samples_from_data(
    sample_data_per_id: dict[str, list[SampleData]],
) -> list[Sample]:
    # generate all samples without any detailed metadata

    samples = [
        Sample(
            sample_id=id,
            scientific_name=id,
            classification=[],
            sample_data=data,
            common_name=None,
        )
        for id, data in sample_data_per_id.items()
    ]

    # try to add metadata if applicable

    add_nucleotide_metadata(samples)
    add_protein_metadata(samples)
    add_language_metadata(samples)

    return samples



import subprocess
from pathlib import Path
from typing import Optional

import click

import phylodata
from phylodata.data_types import get_schema, validate_editable_json
from phylodata.process.change_language import (
    change_language as change_language_handler,
)
from phylodata.process.change_ncbi import (
    change_ncbi as change_ncbi_handler,
)
from phylodata.process.remove_classification import (
    remove_classification as remove_classification_handler,
)


@click.group()
def cli(): ...


@cli.command(help="Process an experiment before uploading it to PhyloData.")
@click.argument("files", nargs=-1, type=click.UNPROCESSED)
def process(files):
    subprocess.run(
        ["streamlit", "run", str(Path(phylodata.__path__[0]) / "process/ui.py"), *files]
    )  # type: ignore


@cli.command(
    help="Validate if a given JSON file contains valid editable PhyloData metadata."
)
@click.argument("file_path", type=click.Path(exists=True))
def validate(file_path: str):
    validate_editable_json(file_path)
    print("File is valid!")


@cli.command(help="Prints the JSON schema for valid PhyloData editable metadata files.")
def schema():
    print(get_schema())


@cli.command(help="Corrects the language of a sample in a metadata file.")
@click.argument("metadata_file", type=click.Path(exists=True))
@click.argument("sample_id", type=str)
@click.argument("language_label", type=str)
def change_language(metadata_file: str, sample_id: str, language_label: str):
    change_language_handler(Path(metadata_file), sample_id, language_label)


@cli.command(
    help="Corrects the NCBI taxon classification of a sample in a metadata file."
)
@click.argument("metadata_file", type=click.Path(exists=True))
@click.argument("sample_id", type=str)
@click.argument("ncbi_taxon_id", type=int)
def change_ncbi(metadata_file: str, sample_id: str, ncbi_taxon_id: str):
    change_ncbi_handler(Path(metadata_file), sample_id, int(ncbi_taxon_id))


@cli.command(
    help="Removes the classification of a sample in a metadata file. If no sample_id is provided, all samples are removed."
)
@click.argument("metadata_file", type=click.Path(exists=True))
@click.argument("sample_id", type=str, required=False)
def remove_classification(metadata_file: str, sample_id: Optional[str] = None):
    remove_classification_handler(Path(metadata_file), sample_id)



class ValidationError(Exception):
    def __init__(self, message: str):
        self.message = message


class BlastError(Exception):
    def __init__(self, message: str):
        self.message = message



import os
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import msgspec
from tqdm import tqdm

from phylodata.data_types import (
    EditablePaperWithExperiment,
    FileType,
    NonEditablePaperWithExperiment,
    PaperWithExperiment,
)
from phylodata.loader.download_file import download_file

EDITABLE_METADATA_FILE = "editable_phylodata_metadata.json"
NON_EDITABLE_METADATA_FILE = "non_editable_phylodata_metadata"


@dataclass
class ExperimentToLoad:
    id: str
    version: Optional[int] = None


def load_experiments(
    experiments_to_load: list[ExperimentToLoad | str],
    directory: Optional[str | Path] = None,
    download_only_preview: Optional[bool] = None,
    files_to_download: Optional[list[FileType]] = None,
    force_download: bool = False,
) -> list[PaperWithExperiment]:
    """Loads multiple PhyloData experiments.

    Args:
        experiments_to_load:
            List of experiment IDs or ExperimentToLoad objects specifying which experiments
            to load. If only the ID strings are given, the latest version will be downloaded.
        directory:
            Path where the experiment files should be stored.
        download_only_preview:
            Whether to only download preview files. This is useful for testing environments.
            This can also be controlled by setting the environment variable
            PHYLODATA_DOWNLOAD_ONLY_PREVIEW to true or false. Defaults to False.
        files_to_download:
                Optional list to restrict the FileTypes downloaded. Defaults to all.
        force_download:
            Whether to re-download files even if they exist locally. Defaults to False.

    Returns:
        A list of PaperWithExperiment objects containing the loaded experiment data.
    """
    experiments = []

    for experiment_to_load in tqdm(experiments_to_load):
        if isinstance(experiment_to_load, str):
            experiment_to_load = ExperimentToLoad(experiment_to_load)

        experiment = load_experiment(
            experiment_id=experiment_to_load.id,
            directory=directory,
            version=experiment_to_load.version,
            download_only_preview=download_only_preview,
            files_to_download=files_to_download,  # type: ignore
            force_download=force_download,
        )
        experiments.append(experiment)

    return experiments


def load_experiment(
    experiment_id: str,
    directory: Optional[str | Path] = None,
    version: Optional[int] = None,
    download_only_preview: Optional[bool] = None,
    files_to_download: Optional[list[str | FileType]] = None,
    force_download: bool = False,
) -> PaperWithExperiment:
    """Loads a PhyloData experiment.

    Args:
        experiment_id:
            The human-readable ID of the experiment to load (e.g. felsenstein-1992-estimating).
        directory:
            Path where the experiment files are stored.
        version:
            The version of the experiment to load. Defaults to latest.
        download_only_preview:
            Whether to only download preview files. This is useful for testing environments.
            This can also be controlled by setting the environment variable
            PHYLODATA_DOWNLOAD_ONLY_PREVIEW to true or false. Defaults to False.
        files_to_download:
            Optional list to restrict the filenames or FileTypes downloaded. Defaults to all.
        force_download:
            Whether to re-download files even if they exist locally. Defaults to False.

    Returns:
        A PaperWithExperiment object containing the experiment data.
    """
    if not directory:
        directory = Path("data") / experiment_id
    if isinstance(directory, str):
        directory = Path(directory) / experiment_id
    directory.mkdir(parents=True, exist_ok=True)

    if download_only_preview is None:
        download_only_preview = (
            os.environ.get("PHYLODATA_DOWNLOAD_ONLY_PREVIEW") == "true"
        )

    paper_with_experiment = _download_experiment(
        experiment_id,
        directory,
        version,
        download_only_preview,
        files_to_download,
        force_download,
    )
    return paper_with_experiment


def _download_experiment(
    experiment_id: str,
    directory: Path,
    version: Optional[int] = None,
    download_only_preview: bool = False,
    files_to_download: Optional[list[str | FileType]] = None,
    force_download: bool = False,
):
    # download and load metadata files

    editable_metadata_file = download_file(
        directory,
        experiment_id,
        EDITABLE_METADATA_FILE,
        version,
        force_download,
    )
    with open(editable_metadata_file, "rb") as handle:
        editable_metadata = msgspec.json.decode(
            handle.read(), type=EditablePaperWithExperiment
        )

    non_editable_metadata_file = download_file(
        directory,
        experiment_id,
        NON_EDITABLE_METADATA_FILE,
        version,
        force_download,
    )
    with open(non_editable_metadata_file, "rb") as handle:
        non_editable_metadata = msgspec.json.decode(
            handle.read(), type=NonEditablePaperWithExperiment
        )

    # merge metadata

    metadata = PaperWithExperiment.from_partial(
        editable_metadata, non_editable_metadata
    )

    # download remaining files if needed

    for file in non_editable_metadata.files:
        if (
            files_to_download
            and file.name not in files_to_download
            and file.type not in files_to_download
        ):
            continue

        if file.type == FileType.PHYLO_DATA_EXPERIMENT:
            continue

        if (
            download_only_preview
            and file.type in [FileType.BEAST2_POSTERIOR_LOGS, FileType.POSTERIOR_TREES]
            and not file.is_preview
        ):
            continue

        download_file(directory, experiment_id, file.name, version, force_download)

    return metadata


def load_experiment_from_local_dir(directory: str | Path) -> PaperWithExperiment:
    """Loads a PhyloData experiment from a local directory.

    Args:
        directory:
            Path to the directory containing the experiment files.

    Returns:
        A PaperWithExperiment object containing the experiment data.
    """
    if isinstance(directory, str):
        directory = Path(directory)

    # load metadata files

    with open(directory / EDITABLE_METADATA_FILE, "rb") as handle:
        editable_metadata = msgspec.json.decode(
            handle.read(), type=EditablePaperWithExperiment
        )

    with open(directory / NON_EDITABLE_METADATA_FILE, "rb") as handle:
        non_editable_metadata = msgspec.json.decode(
            handle.read(), type=NonEditablePaperWithExperiment
        )

    # merge metadata

    experiment = PaperWithExperiment.from_partial(
        editable_metadata, non_editable_metadata
    )
    return experiment



from pathlib import Path

import msgspec

from phylodata.data_types import EditablePaperWithExperiment
from phylodata.process.samples.add_language_metadata import (
    fetch_language_classification,
)


def change_language(metadata_file: Path, sample_id: str, language_label: str):
    """
    Changes the classification of a language sample in a editable metadata file.

    Args:
        metadata_file: The path to the metadata file.
        sample_id: The ID of the sample to correct.
        language_label: The label of the language to correct to.
    """

    with open(metadata_file, "r") as file:
        metadata = msgspec.json.decode(file.read(), type=EditablePaperWithExperiment)

    for sample in metadata.samples:
        if sample.sample_id == sample_id:
            if classification := fetch_language_classification(language_label):
                sample.classification = classification
                sample.scientific_name = classification[0].scientific_name
                sample.common_name = classification[0].common_name
                break

    with open(metadata_file, "wb") as file:
        file.write(msgspec.json.format(msgspec.json.encode(metadata), indent=2))



def is_valid_email(email: str) -> bool:
    """Crudely verifies if an email is valid."""
    if "@" not in email:
        return False

    local, domain = email.split("@")
    if not local or not domain:
        return False

    return True



from datetime import date

import streamlit as st

from phylodata.data_types import (
    EditableExperiment,
    EditablePaper,
    ExperimentType,
    NonEditableExperiment,
)
from phylodata.process.experiment.human_readable_id import generate_human_readable_id
from phylodata.process.module import Module


class ExperimentModule(Module[tuple[EditableExperiment, NonEditableExperiment]]):
    def ui(self):
        with st.container(border=True):
            st.write("""
            ### Experiment

            Optionally, enter some additional information about your experiment. This is useful if you want to register multiple experiments for the same paper.
            """)

            self.title = st.text_input("Experiment title (optional)")
            self.description = st.text_area("Experiment description (optional)")

    def validate(self): ...

    def parse(
        self, paper: EditablePaper
    ) -> tuple[EditableExperiment, NonEditableExperiment]:
        return EditableExperiment(
            title=self.title.strip() or None,
            description=self.description.strip() or None,
            type=ExperimentType.BEAST2_Experiment,
        ), NonEditableExperiment(
            origin="manualUpload",
            version=1,
            upload_date=date.today(),
            human_readable_id=generate_human_readable_id(
                paper.title.strip(),
                paper.year,
                [author.strip() for author in paper.authors],
            ),
        )



from pathlib import Path
from typing import Optional

import msgspec

from phylodata.data_types import EditablePaperWithExperiment


def remove_classification(metadata_file: Path, sample_id: Optional[str] = None):
    """
    Removes the classification of a language sample in a editable metadata file.
    If no sample_id is provided, all samples are removed.

    Args:
        metadata_file: The path to the metadata file.
        sample_id: The ID of the sample to remove the classification from.
    """

    with open(metadata_file, "r") as file:
        metadata = msgspec.json.decode(file.read(), type=EditablePaperWithExperiment)

    for sample in metadata.samples:
        if not sample_id or sample.sample_id == sample_id:
            sample.classification = []
            sample.scientific_name = sample.sample_id
            sample.common_name = None

    with open(metadata_file, "wb") as file:
        file.write(msgspec.json.format(msgspec.json.encode(metadata), indent=2))
